//+------------------------------------------------------------------+
//|                                                CFractalPoint.mqh |
//|                                           Copyright 2016, denkir |
//|                           https://login.mql5.com/ru/users/denkir |
//+------------------------------------------------------------------+
#property copyright "Copyright 2016, denkir"
#property link      "https://login.mql5.com/ru/users/denkir"
//--- include
#include <Arrays\ArrayDouble.mqh>
#include <Arrays\ArrayInt.mqh>
#include <ChartObjects\ChartObjectsChannels.mqh>
#include <Arrays\ArrayObj.mqh>
#include <User\CisNewBar.mqh>
//+------------------------------------------------------------------+
//|                                                     |
//+------------------------------------------------------------------+
enum ENUM_EXTREMUM_TYPE
  {
   EXTREMUM_TYPE_MIN=0, // 
   EXTREMUM_TYPE_MAX=1, // 
  };
//+------------------------------------------------------------------+
//|                                       |
//+------------------------------------------------------------------+
enum ENUM_SET_TYPE
  {
   SET_TYPE_NONE=0,     //  
   SET_TYPE_MINMAX=1,   // --
   SET_TYPE_MAXMIN=2,   // --                       
  };
//+------------------------------------------------------------------+
//|                                                |
//+------------------------------------------------------------------+
enum ENUM_RELEVANT_EXTREMUM
  {
   RELEVANT_EXTREMUM_PREV=0, // 
   RELEVANT_EXTREMUM_LAST=1, // 
  };
//+------------------------------------------------------------------+
//|                                             |
//+------------------------------------------------------------------+
struct SFracData
  {
   double            value; // 
   int               index; //  
   datetime          time;  //  
   //--- 
   void SFracData::SFracData(void)
     {
      value=EMPTY_VALUE;
      index=WRONG_VALUE;
      time=WRONG_VALUE;
     }
  };
//+------------------------------------------------------------------+
//|                                             |
//+------------------------------------------------------------------+
class CFractalPoint : public CObject
  {
   //--- === Data members === --- 
private:
   datetime          m_date;           //   
   double            m_value;          // 
   ENUM_EXTREMUM_TYPE m_extreme_type;  //  
   int               m_idx;            //  ( 0  2)
   //---
   CisNewBar         m_new_bar;            //   

   //--- === Methods === --- 
public:
   //--- /
   void              CFractalPoint(void);
   void              CFractalPoint(datetime _date,double _value,
                                   ENUM_EXTREMUM_TYPE _extreme_type,int _idx);
   void             ~CFractalPoint(void){};
   //--- get-
   datetime          Date(void) const {return m_date;};
   double            Value(void) const {return m_value;};
   ENUM_EXTREMUM_TYPE FractalType(void) const {return m_extreme_type;};
   int               Index(void) const {return m_idx;};
   //--- set-
   void              Date(const datetime _date) {m_date=_date;};
   void              Value(const double _value) {m_value=_value;};
   void              FractalType(const ENUM_EXTREMUM_TYPE extreme_type) {m_extreme_type=extreme_type;};
   void              Index(const int _bar_idx){m_idx=_bar_idx;};
   //--- 
   void              Copy(const CFractalPoint &_source_frac);
   void              Print(void);
  };
//+------------------------------------------------------------------+
//|                                            |
//+------------------------------------------------------------------+
void CFractalPoint::CFractalPoint(void)
  {
   m_date=0;
   m_value=0.;
   m_extreme_type=-1;
   m_idx=WRONG_VALUE;
  };
//+------------------------------------------------------------------+
//|                                 |
//+------------------------------------------------------------------+
void CFractalPoint::CFractalPoint(datetime _date,double _value,
                                  ENUM_EXTREMUM_TYPE _extreme_type,int _idx):
                                  m_date(_date),
                                  m_value(_value),
                                  m_extreme_type(_extreme_type),
                                  m_idx(_idx){};
//+------------------------------------------------------------------+
//|                                                       |
//+------------------------------------------------------------------+
void CFractalPoint::Copy(const CFractalPoint &_source_frac)
  {
   m_date=_source_frac.m_date;
   m_value=_source_frac.m_value;
   m_extreme_type=_source_frac.m_extreme_type;
   m_idx=_source_frac.m_idx;
  };
//+------------------------------------------------------------------+
//|                                                            |
//+------------------------------------------------------------------+
void CFractalPoint::Print(void)
  {
   Print("\n---===    ===---");
   Print(": ",TimeToString(m_date));
   Print(": ",DoubleToString(m_value,_Digits));
   Print(": ",EnumToString(m_extreme_type));
   Print(": ",IntegerToString(m_idx));
  }
//+------------------------------------------------------------------+
//|                                       |
//+------------------------------------------------------------------+
class CFractalSet : protected CArrayObj
  {
   //--- === Data members === --- 
private:
   ENUM_SET_TYPE     m_set_type;           //   
   int               m_fractal_num;        //   
   int               m_fractals_ha;        //    
   CisNewBar         m_new_bar;            //   
   bool              m_is_init;            //  
   //---  
   int               m_prev_frac_num;      //  
   int               m_bars_beside;        //  /  
   int               m_bars_between;       //     
   ENUM_RELEVANT_EXTREMUM m_rel_frac;      //  
   int               m_line_wid;           //  
   bool              m_to_log;             //  ?
   //---
   datetime          m_last_frac_date;     // /    
   //--- === Methods === --- 
public:
   //--- /
   void              CFractalSet(void);
   void              CFractalSet(const CFractalSet &_src_frac_set);
   void             ~CFractalSet(void){};
   //---
   void              operator=(const CFractalSet &_src_frac_set);
   //--- 
   bool              Init(
                          int _prev_frac_num,
                          int _bars_beside,
                          int _bars_between=0,
                          ENUM_RELEVANT_EXTREMUM _rel_frac=RELEVANT_EXTREMUM_PREV,
                          int _line_wid=3,
                          bool _to_log=true
                          );
   void              Calculate(
                               const double &_up_frac_buffer[],
                               const double &_down_frac_buffer[],
                               const datetime &_time_arr[],
                               double &_up_buffer[],
                               double &_down_buffer[],
                               double &_new_ch_buffer[]
                               );
   //--- 
   bool              IsInit(void) const {return m_is_init;};
   ENUM_SET_TYPE     GetTypeOfSet(void) const {return m_set_type;};
   //---
private:
   int               CheckSet(const SFracData &_fractals[]);
   void              SetTypeOfSet(const ENUM_SET_TYPE _set_type) {m_set_type=_set_type;};
   bool              Crop(const uint _num_to_crop);
   void              BubbleSort(void);
  };
//+------------------------------------------------------------------+
//|                                                       |
//+------------------------------------------------------------------+
void CFractalSet::CFractalSet(void)
  {
   m_set_type=WRONG_VALUE;
   m_fractal_num=3;
   m_prev_frac_num=m_bars_beside=m_bars_between=m_line_wid=0;
   m_rel_frac=WRONG_VALUE;
   m_is_init=false;
   m_to_log=false;
   m_last_frac_date=0;
  };
//+------------------------------------------------------------------+
//|                                                     |
//+------------------------------------------------------------------+
bool CFractalSet::Init(
                       int _prev_frac_num,
                       int _bars_beside,
                       int _bars_between=0,
                       ENUM_RELEVANT_EXTREMUM _rel_frac=RELEVANT_EXTREMUM_PREV,
                       int _line_wid=3,
                       bool _to_log=true
                       )
  {

//---     
   if((_prev_frac_num<0) || (_prev_frac_num>3))
     {
      Print("     !");
      return false;
     }
//---     
   if(_bars_beside<1)
     {
      Print("        !");
      return false;
     }

//---  
   if(!this.FreeMode())
      this.FreeMode(true);
//--- -
   m_prev_frac_num=_prev_frac_num;
   m_bars_between=_bars_between;
   m_rel_frac=_rel_frac;
   m_line_wid=_line_wid;
   m_bars_beside=_bars_beside;
   m_to_log=_to_log;
   if(m_to_log)
     {
      Print("    \"EquidistantChannels\".");
     }
//---       
   m_new_bar.SetLastBarTime(1);
   m_is_init=true;
//---
   return true;
  }
//+------------------------------------------------------------------+
//|                                              |
//+------------------------------------------------------------------+
void CFractalSet::Calculate(
                            const double &_up_frac_buffer[],
                            const double &_down_frac_buffer[],
                            const datetime &_time_arr[],
                            double &_up_buffer[],
                            double &_down_buffer[],
                            double &_new_ch_buffer[]
                            )
  {
//---  -
   ArraySetAsSeries(_up_frac_buffer,true);
   ArraySetAsSeries(_down_frac_buffer,true);
   ArraySetAsSeries(_time_arr,true);
   ArraySetAsSeries(_up_buffer,true);
   ArraySetAsSeries(_down_buffer,true);
   ArraySetAsSeries(_new_ch_buffer,true);
//---   
   if(m_new_bar.isNewBar(_time_arr[0]))
     {
      int frac_num=this.Total();
      //---      
      if(m_prev_frac_num==0)
         if(m_last_frac_date==0)
           {
            m_last_frac_date=_time_arr[0];
            return;
           }
      //---   ?
      bool is_history_add=false;
      if(m_prev_frac_num>0)
         if(!this.IsSorted(1))
            is_history_add=true;
      //---  
      for(int bar=0;bar<ArraySize(_up_frac_buffer);bar++)
        {
         SFracData fracs[2]; //    : [0]-, [1]-               
         double curr_up_val,curr_down_val;
         curr_up_val=_up_frac_buffer[bar];
         curr_down_val=_down_frac_buffer[bar];
         datetime curr_bar_time=_time_arr[bar];
         bool is_new_down=false,is_new_up=false;
         //---    
         if((curr_down_val<DBL_MAX) && (curr_down_val>0.))
           {
            is_new_down=true;
            fracs[0].index=bar;
            fracs[0].time=curr_bar_time;
           }
         if((curr_up_val<DBL_MAX) && (curr_up_val>0.))
           {
            is_new_up=true;
            fracs[1].index=bar;
            fracs[1].time=curr_bar_time;
           }
         //---    
         if(is_new_down || is_new_up)
           {
            bool to_add=false;
            //---   
            frac_num=this.Total();
            //---    :            
            //--- 1)    
            if(is_history_add)
              {
               //---   
               if(frac_num<m_prev_frac_num)
                  //---     
                  if((curr_bar_time<m_last_frac_date) || (m_last_frac_date==0))
                    {
                     if(m_last_frac_date==0)
                        m_last_frac_date=curr_bar_time;
                     to_add=true;
                    }
              }
            //--- 2)    
            else
              {
               //---         
               if(curr_bar_time>m_last_frac_date)
                 {
                  m_last_frac_date=curr_bar_time;
                  to_add=true;
                 }
              }
            //---  
            if(to_add)
              {
               if(m_to_log)
                 {
                  if(is_new_down)
                    {
                     Print("\n---==   ==---");
                     PrintFormat(": %."+IntegerToString(_Digits)+"f",curr_down_val);
                    }
                  if(is_new_up)
                    {
                     Print("\n---==   ==---");
                     PrintFormat(": %."+IntegerToString(_Digits)+"f",curr_up_val);
                    }
                  PrintFormat(": %s",TimeToString(curr_bar_time));
                  PrintFormat(" : %d",bar);
                 }
               //---
               fracs[0].value=curr_down_val;
               fracs[1].value=curr_up_val;
               //---   
               frac_num=this.CheckSet(fracs);
               //---   
               if(frac_num==3)
                 {
                  //---  ,    
                  _new_ch_buffer[0]=1.;
                  break;
                 }
              }
           }
        }
      //--- 
      if(is_history_add)
         if(frac_num>0)
            if(!this.IsSorted(1))
              {
               this.BubbleSort();
               //--- 
               if(!this.IsSorted(1))
                 {
                  Print("   .");
                  this.Clear();
                  return;
                 }
               m_prev_frac_num=0; //      
               //---          
               if(m_to_log)
                  PrintFormat("  : %d",frac_num);
               //}
              }
      //---  
      if(frac_num==3)
        {
         //---  
         ArrayInitialize(_up_buffer,0.);     //   
         ArrayInitialize(_down_buffer,0.);   //   
         //--- - 
         double prices[3];
         datetime times[3];
         ArrayInitialize(times,0);
         ArrayInitialize(prices,0.);
         bool is_1_point=false;
         ENUM_SET_TYPE curr_set_type=this.GetTypeOfSet();
         if(curr_set_type<=SET_TYPE_NONE)
           {
            Print("   :      !");
            return;
           }
         for(int idx=0;idx<ArraySize(prices);idx++)
           {
            //---   
            CFractalPoint *ptr_curr_frac=this.At(idx);
            if(CheckPointer(ptr_curr_frac)!=POINTER_DYNAMIC)
              {
               Print("       !");
               return;
              }
            ENUM_EXTREMUM_TYPE curr_frac_type=ptr_curr_frac.FractalType();
            datetime curr_point_time=ptr_curr_frac.Date();
            double curr_point_price=ptr_curr_frac.Value();
            //---  
            if(curr_set_type==SET_TYPE_MINMAX)
              {
               if(curr_frac_type==EXTREMUM_TYPE_MIN)
                 {
                  //--- 1- 
                  if(!is_1_point)
                    {
                     times[0]=curr_point_time;
                     prices[0]=curr_point_price;
                     is_1_point=true;
                    }
                  //--- 2- 
                  else
                    {
                     times[1]=curr_point_time;
                     prices[1]=curr_point_price;
                    }
                 }
               //---  
               else
                 {
                  times[2]=curr_point_time;
                  prices[2]=curr_point_price;
                 }
              }
            else if(curr_set_type==SET_TYPE_MAXMIN)
              {
               if(curr_frac_type==EXTREMUM_TYPE_MAX)
                 {
                  //--- 1-  
                  if(!is_1_point)
                    {
                     times[0]=curr_point_time;
                     prices[0]=curr_point_price;
                     is_1_point=true;
                    }
                  //--- 2- 
                  else
                    {
                     times[1]=curr_point_time;
                     prices[1]=curr_point_price;
                    }
                 }
               //---  
               else
                 {
                  times[2]=curr_point_time;
                  prices[2]=curr_point_price;
                 }
              }
           }
         //--- 1)  
         //---  
         int first_date_idx=ArrayMinimum(times);
         if(first_date_idx<0)
           {
            Print("   !");
            return;
           }
         datetime first_point_date=times[first_date_idx];
         //---  
         datetime last_point_date=_time_arr[0];
         //--- 2)  
         //--- 2.1  
         //---   1-  2- 
         datetime bars_dates[];
         int bars_between=CopyTime(_Symbol,_Period,
                                   times[0],times[1],bars_dates
                                   );
         if(bars_between<2)
           {
            Print("     !");
            return;
           }
         bars_between-=1;
         //---  
         double price_differential=MathAbs(prices[0]-prices[1]);
         //---   (   1- )
         double price_speed=price_differential/bars_between;
         //---  
         bool is_up=(prices[0]<prices[1]);
         //--- 2.2   1-  3-   
         if(times[0]!=times[2])
           {
            datetime start,end;
            start=times[0];
            end=times[2];
            //---  3-    1-
            bool is_3_point_earlier=false;
            if(times[2]<times[0])
              {
               start=times[2];
               end=times[0];
               is_3_point_earlier=true;
              }
            //---   1-  3- 
            int bars_between_1_3=CopyTime(_Symbol,_Period,
                                          start,end,bars_dates
                                          );
            if(bars_between_1_3<2)
              {
               Print("     !");
               return;
              }
            bars_between_1_3-=1;
            //---   
            if(is_up)
              {
               //---  3-   
               if(is_3_point_earlier)
                  prices[0]-=(bars_between_1_3*price_speed);
               else
                  prices[2]-=(bars_between_1_3*price_speed);
              }
            //---    
            else
              {
               //---  3-   
               if(is_3_point_earlier)
                  prices[0]+=(bars_between_1_3*price_speed);
               else
                  prices[2]+=(bars_between_1_3*price_speed);
              }
           }
         //--- 2.3   2-  
         if(times[1]<last_point_date)
           {
            datetime dates_for_last_bar[];
            //---   2-    
            bars_between=CopyTime(_Symbol,_Period,times[1],last_point_date,dates_for_last_bar);
            if(bars_between<2)
              {
               Print("     !");
               return;
              }
            bars_between-=1;
            //---   
            if(is_up)
               prices[1]+=(bars_between*price_speed);
            //---    
            else
               prices[1]-=(bars_between*price_speed);
           }
         //---    
         times[0]=times[2]=first_point_date;
         times[1]=last_point_date;
         //---    
         int bars_len=CopyTime(_Symbol,_Period,
                               first_point_date,last_point_date,bars_dates
                               );
         //---  --
         if(curr_set_type==SET_TYPE_MINMAX)
           {
            //---  
            if(is_up)
              {
               for(int idx=bars_len-1,jdx=0;idx>=0;idx--,jdx++)
                 {
                  //---   
                  _up_buffer[idx]=prices[2]+jdx*price_speed;
                  //---   
                  _down_buffer[idx]=prices[0]+jdx*price_speed;
                 }
              }
            //---  
            else
              {
               for(int idx=bars_len-1,jdx=0;idx>=0;idx--,jdx++)
                 {
                  //---   
                  _up_buffer[idx]=prices[2]-jdx*price_speed;
                  //---   
                  _down_buffer[idx]=prices[0]-jdx*price_speed;
                 }
              }
           }
         //---  --
         else if(curr_set_type==SET_TYPE_MAXMIN)
           {
            //---  
            if(is_up)
              {
               for(int idx=bars_len-1,jdx=0;idx>=0;idx--,jdx++)
                 {
                  //---   
                  _up_buffer[idx]=prices[0]+jdx*price_speed;
                  //---   
                  _down_buffer[idx]=prices[2]+jdx*price_speed;
                 }
              }
            //---  
            else
              {
               for(int idx=bars_len-1,jdx=0;idx>=0;idx--,jdx++)
                 {
                  //---   
                  _up_buffer[idx]=prices[0]-jdx*price_speed;
                  //---   
                  _down_buffer[idx]=prices[2]-jdx*price_speed;
                 }
              }
           }
        }
     }
//---   
   ArraySetAsSeries(_up_frac_buffer,false);
   ArraySetAsSeries(_down_frac_buffer,false);
   ArraySetAsSeries(_time_arr,false);
   ArraySetAsSeries(_up_buffer,false);
   ArraySetAsSeries(_down_buffer,false);
   ArraySetAsSeries(_new_ch_buffer,false);
  }
//+------------------------------------------------------------------+
//|                                                     |
//+------------------------------------------------------------------+
int CFractalSet::CheckSet(const SFracData &_fractals[])
  {
//---      
   CArrayObj temp_add_set;
   for(int idx=0;idx<ArraySize(_fractals);idx++)
      if((_fractals[idx].value<DBL_MAX) && (_fractals[idx].value>0.))
        {
         //---    
         CFractalPoint *ptr_new_fractal=new CFractalPoint;
         if(CheckPointer(ptr_new_fractal)!=POINTER_DYNAMIC)
            return -1;
         //---
         ENUM_EXTREMUM_TYPE new_fractal_type=WRONG_VALUE;
         //---     
         //--- 1) 
         ptr_new_fractal.Date(_fractals[idx].time);
         //--- 2) 
         double frac_pr=_fractals[idx].value;
         ptr_new_fractal.Value(frac_pr);
         //--- 3) 
         if(idx==1)
            new_fractal_type=EXTREMUM_TYPE_MAX;
         else
            new_fractal_type=EXTREMUM_TYPE_MIN;
         ptr_new_fractal.FractalType(new_fractal_type);
         //---   
         if(!temp_add_set.Add(ptr_new_fractal))
           {
            Print("   !");
            delete ptr_new_fractal;
            return -1;
           }
        }
//---       
   int frac_num_to_add=temp_add_set.Total();
   if(frac_num_to_add<1)
      return -1;
//---
   bool is_emptied=false; //  ?
   int curr_fractal_num=0;
//---    
   for(int frac_idx=0;frac_idx<frac_num_to_add;frac_idx++)
     {
      CFractalPoint *ptr_temp_frac=temp_add_set.At(frac_idx);
      if(CheckPointer(ptr_temp_frac)!=POINTER_DYNAMIC)
        {
         Print("       !");
         return -1;
        }
      //---         
      if(m_bars_between>0)
        {
         curr_fractal_num=this.Total();
         if(curr_fractal_num>0)
           {
            CFractalPoint *ptr_prev_frac=this.At(curr_fractal_num-1);
            if(CheckPointer(ptr_prev_frac)!=POINTER_DYNAMIC)
              {
               Print("      !");
               return -1;
              }
            datetime time1,time2;
            time1=ptr_prev_frac.Date();
            time2=ptr_temp_frac.Date();
            //---   
            datetime bars_dates[];
            int bars_between=CopyTime(_Symbol,_Period,
                                      time1,time2,bars_dates
                                      );
            if(bars_between<0)
              {
               Print("     !");
               return -1;
              }
            bars_between-=2;
            //---    
            if(bars_between>=0)
               //---     
               if(bars_between<m_bars_between)
                 {
                  bool to_delete_frac=false;
                  if(m_to_log)
                     Print("  .    .");
                  //---    
                  if(m_rel_frac==RELEVANT_EXTREMUM_PREV)
                    {
                     datetime curr_frac_date=time2;
                     //---   
                     if(m_is_init)
                       {
                        continue;
                       }
                     //---    
                     else
                       {
                        //---   
                        to_delete_frac=true;
                        curr_frac_date=time1;
                       }
                     if(m_to_log)
                       {
                        PrintFormat("   : %s",
                                    TimeToString(curr_frac_date));
                       }
                    }
                  //---    
                  else
                    {
                     datetime curr_frac_date=time1;
                     //---   
                     if(m_is_init)
                       {
                        //---   
                        to_delete_frac=true;
                       }
                     //---    
                     else
                       {
                        curr_frac_date=time2;
                       }
                     if(m_to_log)
                        PrintFormat("   : %s",
                                    TimeToString(curr_frac_date));
                     if(curr_frac_date==time2)
                        continue;

                    }
                  //---   
                  if(to_delete_frac)
                    {
                     if(!this.Delete(curr_fractal_num-1))
                       {
                        Print("     !");
                        return -1;
                       }
                    }
                 }
           }
        }
      //---       - 
      CFractalPoint *ptr_new_fractal=new CFractalPoint;
      if(CheckPointer(ptr_new_fractal)==POINTER_DYNAMIC)
        {
         ptr_new_fractal.Copy(ptr_temp_frac);
         if(!this.Add(ptr_new_fractal))
           {
            Print("      !");
            delete ptr_new_fractal;
            return -1;
           }
         //---  
         ptr_new_fractal.Index(this.Total()-1);
        }
     }
//---  
   curr_fractal_num=this.Total();
//---    
   if(curr_fractal_num>m_fractal_num)
     {
      uint num_to_crop=curr_fractal_num-m_fractal_num;
      //---  
      if(!this.Crop(num_to_crop))
        {
         Print("     !");
         return -1;
        }
      //---     
      curr_fractal_num=this.Total();
     }
//---   
   if(curr_fractal_num==m_fractal_num)
     {
      //---   
      int min_cnt,max_cnt; // 
      min_cnt=max_cnt=0;
      for(int frac_idx=0;frac_idx<curr_fractal_num;frac_idx++)
        {
         //---  
         CFractalPoint *ptr_curr_frac=this.At(frac_idx);
         if(CheckPointer(ptr_curr_frac)!=POINTER_DYNAMIC)
           {
            Print("      !");
            return -1;
           }
         //---  
         ENUM_EXTREMUM_TYPE curr_frac_type=ptr_curr_frac.FractalType();
         if(curr_frac_type==EXTREMUM_TYPE_MIN)
            min_cnt++;
         else if(curr_frac_type==EXTREMUM_TYPE_MAX)
            max_cnt++;
        }
      //---  2   1 
      if((min_cnt==2) && (max_cnt==1))
         this.SetTypeOfSet(SET_TYPE_MINMAX);
      //---  1   2 
      else if((min_cnt==1) && (max_cnt==2))
         this.SetTypeOfSet(SET_TYPE_MAXMIN);
      //---      ,    
      else
        {
         //---    -  
         if(m_is_init)
           {
            if(!this.Crop(1))
              {
               Print("  1-    !");
               this.Clear();
               return false;
              }
            if(m_to_log)
               Print(" 1-    .");
           }
         //---      -  
         else
           {
            if(!this.Delete(curr_fractal_num-1))
              {
               Print("      !");
               this.Clear();
               return false;
              }
            if(m_to_log)
               Print("     .");
           }
        }
     }
//---
   return this.Total();
  }
//+------------------------------------------------------------------+
//|                                                    |
//+------------------------------------------------------------------+
bool CFractalSet::Crop(const uint _num_to_crop)
  {
//---      
   if(_num_to_crop<1)
      return false;
//---
   if(!this.DeleteRange(0,_num_to_crop-1))
     {
      Print("   !");
      return false;
     }
//---
   return true;
  }
//+------------------------------------------------------------------+
//|                                             |
//+------------------------------------------------------------------+
void CFractalSet::BubbleSort(void)
  {
   m_sort_mode=-1;
//---
   uint arr_size=this.Total();
   for(uint passes=0;passes<arr_size-1;passes++)
      for(uint j=0;j<(arr_size-passes-1);j++)
        {
         CFractalPoint *ptr_p1=this.At(j);
         CFractalPoint *ptr_p2=this.At(j+1);
         //---   ( )
         if(ptr_p1.Date()>ptr_p2.Date())
           {
            CFractalPoint *ptr_temp_p=new CFractalPoint;
            if(CheckPointer(ptr_temp_p)==POINTER_DYNAMIC)
              {
               ptr_temp_p.Copy(ptr_p2);
               if(!this.Insert(ptr_temp_p,j))
                  return;
               if(!this.Delete(j+2))
                  return;
              }
            else
               return;
           }
        }
   m_sort_mode=1;
  }
//+------------------------------------------------------------------+
